@@ -1,9 +1,9 @@
 """
-TkinterWeb v3.16
+TkinterWeb v3.23
 This is a wrapper for the Tkhtml3 widget from http://tkhtml.tcl.tk/tkhtml.html, 
 which displays styled HTML documents in Tkinter.
 
-Copyright (c) 2022 Andereoo
+Copyright (c) 2024 Andereoo
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -36,16 +36,19 @@ except (ImportError, ModuleNotFoundError):
     # Give useful troubleshooting information as a popup, as most bundled applications don't have a visible console
     # Also print the message in case something is wrong with the Tkinter installation as well
     error_message = "ModuleNotFoundError: The files required to run TkinterWeb could not be found. \
-This usually occurs when bundling TkinterWeb into an app without forcing the application maker to include all nessessary files. \
-See https://github.com/Andereoo/TkinterWeb/blob/main/tkinterweb/docs/FAQ.md for more information."
-    sys.stdout.write("{}\n\nError: {}\n\nIf you think this is a bug, please file a bug report at https://github.com/Andereoo/TkinterWeb.".format(error_message, traceback.format_exc()))
+This usually occurs when bundling TkinterWeb into an app without forcing the application maker to include all nessessary files or when some of TkinterWeb's dependencies are not installed. \
+See https://github.com/Andereoo/TkinterWeb/blob/main/tkinterweb/docs/FAQ.md for more information. \n\n\
+Error code: {} \n\
+If you think this is a bug, please file a bug report at https://github.com/Andereoo/TkinterWeb.".format(traceback.format_exc())
+    sys.stdout.write(error_message)
     
     try:
         import tkinter as tk
         from tkinter import messagebox
-    except ImportError as e:
+    except ImportError:
         import Tkinter as tk
         import tkMessageBox as messagebox
+        
     root = tk.Tk()
     root.withdraw()
     message = messagebox.showerror("Fatal Error Encountered", error_message) #for older versions of pyinstaller, windowed app may crash without any message of any kind
@@ -54,7 +57,7 @@ See https://github.com/Andereoo/TkinterWeb/blob/main/tkinterweb/docs/FAQ.md for
 
 __title__ = 'TkinterWeb'
 ___author__ = "Andereoo"
-__version__ = '3.16'
+__version__ = '3.23'
 __all__ = ['HtmlFrame', 'HtmlLabel', 'TkinterWeb', 'Demo']
 
 
