@@ -1,3 +1,29 @@
+"""
+TkinterWeb v3.23
+This is a wrapper for the Tkhtml3 widget from http://tkhtml.tcl.tk/tkhtml.html, 
+which displays styled HTML documents in Tkinter.
+
+Copyright (c) 2024 Andereoo
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
+
 from bindings import TkinterWeb
 from utilities import (AutoScrollbar, StoppableThread, cachedownload, download,
                    notifier, currentpath, threadname)
@@ -18,7 +44,7 @@ except ImportError:
 
 
 class HtmlFrame(ttk.Frame):
-    def __init__(self, master, messages_enabled=True, vertical_scrollbar="auto", horizontal_scrollbar=False, **kw):
+    def __init__(self, master, messages_enabled=True, vertical_scrollbar="auto", horizontal_scrollbar=False, scroll_overflow=None, **kw):
         ttk.Frame.__init__(self, master, **kw)
 
         if messages_enabled:
@@ -42,6 +68,8 @@ class HtmlFrame(ttk.Frame):
             vsb.bind("<MouseWheel>", self.scroll)
             vsb.bind("<Button-4>", self.scroll_x11)
             vsb.bind("<Button-5>", self.scroll_x11)
+            html.bind("<Button-4>", self.overflow_scroll_x11)
+            html.bind("<Button-5>", self.overflow_scroll_x11)
             self.bind_class("{0}.document".format(html),
                             "<MouseWheel>", self.scroll)
             self.bind_class(html.scrollable_node_tag,
@@ -67,10 +95,16 @@ class HtmlFrame(ttk.Frame):
             html.configure(xscrollcommand=hsb.set)
             hsb.grid(row=1, column=0, sticky=tk.NSEW)
 
+        self.bind("<Leave>", html.on_leave)
+        self.bind("<Enter>", html.on_mouse_motion)
+
         # state and settings variables
         self.master = master
+        self.scroll_overflow = scroll_overflow
         self.current_url = ""
         self.cursor = ""
+        self.accumulated_styles = []
+        self.waiting_for_reset = False
         self.image_count = 0
         self.image = None
         self.thread_in_progress = None
@@ -89,7 +123,7 @@ class HtmlFrame(ttk.Frame):
         self.html.done_loading_func = self.done_loading
 
         self.message_func(
-            "Welcome to TkinterWeb 3.16! \nhttps://github.com/Andereoo/TkinterWeb")
+            "Welcome to TkinterWeb 3.23! \nhttps://github.com/Andereoo/TkinterWeb")
 
         self.message_func(
             "Debugging messages are enabled. \nUse the parameter `messages_enabled = False` when calling HtmlFrame() to disable these messages.")
@@ -97,26 +131,53 @@ class HtmlFrame(ttk.Frame):
         self.columnconfigure(0, weight=1)
         self.rowconfigure(0, weight=1)
 
-    def load_website(self, website_url, decode=None, force=False):
+        #Redirected commands
+        self.select_all = self.html.select_all
+        self.bind = self.html.bind
+        self.set_zoom = self.html.set_zoom
+        self.get_zoom = self.html.get_zoom
+        self.set_fontscale = self.html.set_fontscale
+        self.get_fontscale = self.html.get_fontscale
+        self.set_parsemode = self.html.set_parsemode
+        self.get_parsemode = self.html.get_parsemode
+        self.resolve_url = self.html.resolve_url
+
+        self.yview = self.html.yview
+        self.yview_moveto = self.html.yview_moveto
+        self.yview_scroll = self.html.yview_scroll
+        
+    def yview_toelement(self, selector, index=0):
+        "Find an element that matches a given CSS selectors and scroll to it"
+        nodes = self.html.search(selector)
+        if nodes:
+            try:
+                self.html.yview(nodes[index])
+            except IndexError:
+                pass
+
+    def load_website(self, website_url, decode=None, force=False, insecure=False):
         "Load a website from the specified URL"
         if (not website_url.startswith("https://")) and (not website_url.startswith("http://")) and (not website_url.startswith("about:")):
             website_url = "http://" + str(website_url)
-        self.load_url(website_url, decode, force)
+        self.load_url(website_url, decode, force, insecure)
 
-    def load_file(self, file_url, decode=None, force=False):
+    def load_file(self, file_url, decode=None, force=False, insecure=False):
         "Load a locally stored file from the specified path"
-        file_url = str(file_url).split("/", 1)[1]
-        if not file_url.startswith("/"):
-            file_url = "/" + file_url
-        file_url = "file://" + file_url
-        self.load_url(file_url, decode, force)
+        if not file_url.startswith("file://"):
+            if platform.system() == "Windows" and not file_url.startswith("/"):
+                file_url = "file:///" + str(file_url)
+            else:
+                file_url = "file://" + str(file_url)
+        self.load_url(file_url, decode, force, insecure)
 
-    def load_url(self, url, decode=None, force=False):
+    def load_url(self, url, decode=None, force=False, insecure=False):
         """Load a website (https:// or http://) or a file (file://) from the specified URL.
         We use threading here to prevent the GUI from freezing while fetching the website.
         Technically Tkinter isn't threadsafe and will crash when doing this, but under certain circumstances we can get away with it.
         As long as we do not use the .join() method and no errors are raised in the mainthread, we should be okay.
         """
+        self.waiting_for_reset = True
+
         #Workaround for Bug #40, where urllib.urljoin constructs improperly formatted urls on Linux when url starts with file:///
         if not url.startswith("file://///"):
             url = url.replace("file:////", "file:///")
@@ -125,11 +186,11 @@ class HtmlFrame(ttk.Frame):
             self.thread_in_progress.stop()
         if self.html.max_thread_count >= 1:
             thread = StoppableThread(target=self.continue_loading, args=(
-                url,), kwargs={"decode": decode, "force": force})
+                url,), kwargs={"decode": decode, "force": force, "insecure": insecure})
             self.thread_in_progress = thread
             thread.start()
         else:
-            self.continue_loading(url, decode=decode, force=force)
+            self.continue_loading(url, decode=decode, force=force, insecure=insecure)
 
     def load_form_data(self, url, data, method="GET", decode=None):
         "Load a webpage using form data"
@@ -137,14 +198,15 @@ class HtmlFrame(ttk.Frame):
             self.thread_in_progress.stop()
         if self.html.max_thread_count >= 1:
             thread = StoppableThread(
-                target=self.continue_loading, args=(url, data, method, decode,))
+                target=self.continue_loading, args=(url, data, method, decode))
             self.thread_in_progress = thread
             thread.start()
         else:
             self.continue_loading(url, data, method, decode)
 
-    def continue_loading(self, url, data="", method="GET", decode=None, force=False):
+    def continue_loading(self, url, data="", method="GET", decode=None, force=False, insecure=False):
         "Finish loading urls and handle URI fragments"
+
         self.html.downloading_resource_func()
         self.url_change_func(url)
         try:
@@ -155,18 +217,24 @@ class HtmlFrame(ttk.Frame):
                 url = str(url) + str(data)
 
             # if url is different than the current one, load the new site.
-            if force or (method == "POST") or (urldefrag(url)[0] != urldefrag(self.current_url)[0]):
+            if force or (method == "POST") or (self.skim(urldefrag(url)[0]) != self.skim(urldefrag(self.current_url)[0])):
                 self.message_func("Connecting to {0}.".format(parsed.netloc))
+                if insecure:
+                    self.message_func("WARNGING: Using insecure HTTPS session")
                 if (parsed.scheme == "file") or (not self.html.caches_enabled):
                     data, newurl, filetype = download(
-                        url, data, method, decode)
+                        url, data, method, decode, insecure)
                 else:
                     data, newurl, filetype = cachedownload(
-                        url, data, method, decode)
+                        url, data, method, decode, insecure)
                 if threadname().isrunning():
                     self.url_change_func(newurl)
                     if "image" in filetype:
                         image, error = newimage(data, "_htmlframe_img_{}_{}_".format(id(self), self.image_count), filetype, self.html.image_inversion_enabled)
+                        if error:
+                            self.html.image_setup_func(url, False)
+                        else:
+                            self.html.image_setup_func(url, True)
                         self.load_html("<img style='max-width:100%' src='replace:{}'></img".format(image))
                         self.image_count += 1
                         self.image = image
@@ -176,11 +244,15 @@ class HtmlFrame(ttk.Frame):
             else:
                 # if no requests need to be made, we can signal that the page is done loading
                 self.html.done_loading_func()
+                self.finish_css()
+
             # handle URI fragments
             frag = parsed.fragment
             if frag:
                 #self.html.tk.call(self.html._w, "_force")
+                self.html.update()
                 try:
+                    frag = ''.join(char for char in frag if char.isalnum() or char in ("-", "_"))
                     node = self.html.search("[id=%s]" % frag)
                     if node:
                         self.html.yview(node)
@@ -198,6 +270,9 @@ class HtmlFrame(ttk.Frame):
 
         self.thread_in_progress = None
 
+    def skim(self, url):
+        return url.replace("/", "")
+
     def stop(self):
         "Stop loading a page"
         if self.thread_in_progress:
@@ -206,42 +281,10 @@ class HtmlFrame(ttk.Frame):
         self.url_change_func(self.current_url)
         self.done_loading()
 
-    def bind(self, *args, **kwargs):
-        "Redirect bindings"
-        self.html.bind(*args, **kwargs)
-
     def done_loading(self):
         self.in_progress = False
         self.done_loading_func()
 
-    def set_zoom(self, multiplier):
-        "Set the zoom multiplier"
-        self.html.zoom(float(multiplier))
-
-    def get_zoom(self):
-        "Get the zoom multiplier"
-        return self.html.zoom(None)
-
-    def set_fontscale(self, multiplier):
-        "Set the fontsize multiplier"
-        self.html.fontscale(float(multiplier))
-
-    def get_fontscale(self):
-        "Get the fontsize multiplier"
-        return self.html.fontscale(None)
-
-    def set_parsemode(self, mode):
-        "Set the page render mode"
-        self.html.parsemode(mode)
-
-    def get_parsemode(self):
-        "Get the page render mode"
-        return self.html.parsemode(None)
-     
-    def resolve_url(self, url):
-        "Get full url from partial url"
-        return self.html.resolve_url(url)
-
     def on_link_click(self, function):
         "Allows for handling link clicks"
         self.html.link_click_func = function
@@ -266,6 +309,10 @@ class HtmlFrame(ttk.Frame):
         "Alllows for handling the finishing of all outstanding requests"
         self.done_loading_func = function
 
+    def on_image_setup(self, function):
+        "Callback for image loading"
+        self.html.image_setup_func = function
+
     def on_downloading_resource(self, function):
         "Allows for handling resource downloads"
         self.html.downloading_resource_func = function
@@ -334,9 +381,25 @@ class HtmlFrame(ttk.Frame):
         self.html.image_inversion_enabled = invert_images
         self.html.update_default_style()
 
-    def find_text(self, searchtext, select=1, ignore_case=True, highlight_all=True):
+    def copy_settings(self, html):
+        self.set_message_func(html.message_func)
+        self.set_recursive_hover_depth(html.recursive_hovering_count)
+        self.set_maximum_thread_count(html.max_thread_count)
+        self.ignore_invalid_images(html.ignore_invalid_images)
+        self.enable_stylesheets(html.stylesheets_enabled)
+        self.enable_images(html.images_enabled)
+        self.enable_forms(html.forms_enabled)
+        self.enable_objects(html.objects_enabled)
+        self.enable_caches(html.caches_enabled)
+        self.set_parsemode(html.get_parsemode())
+
+    def find_text(self, searchtext, select=1, ignore_case=True, highlight_all=True, detailed=False):
         "Search for and highlight specific text"
-        return self.html.find_text(searchtext, select, ignore_case, highlight_all)
+        nmatches, selected, matches = self.html.find_text(searchtext, select, ignore_case, highlight_all)
+        if detailed:
+            return nmatches, selected, matches
+        else:
+            return nmatches
 
     def change_cursor(self, cursor):
         "Handle cursor changes"
@@ -358,10 +421,10 @@ class HtmlFrame(ttk.Frame):
     def get_currently_hovered_node_tag(self):
         "Get the tag of the HTML element the mouse pointer is currently over"
         try:
-            tag = self.html.get_node_tag(self.html.currently_hovered_node)
+            tag = self.html.get_node_tag(self.html.current_node)
             if tag == "":
                 tag = self.html.get_node_tag(
-                    self.html.get_node_parent(self.html.currently_hovered_node))
+                    self.html.get_node_parent(self.html.current_node))
         except tk.TclError:
             tag = ""
         return tag
@@ -369,10 +432,10 @@ class HtmlFrame(ttk.Frame):
     def get_currently_hovered_node_text(self):
         "Get the text content of the HTML element the mouse pointer is currently over"
         try:
-            text = self.html.get_node_text(self.html.currently_hovered_node)
+            text = self.html.get_node_text(self.html.current_node)
             if text == "":
                 text = self.html.get_node_text(
-                    self.html.get_node_parent(self.html.currently_hovered_node))
+                    self.html.get_node_parent(self.html.current_node))
         except tk.TclError:
             text = ""
         return text
@@ -385,10 +448,10 @@ class HtmlFrame(ttk.Frame):
         """
         try:
             attr = self.html.get_node_attribute(
-                self.html.currently_hovered_node, attribute)
+                self.html.current_node, attribute)
             if attr == "":
                 attr = self.html.get_node_attribute(self.html.get_node_parent(
-                    self.html.currently_hovered_node), attribute)
+                    self.html.current_node), attribute)
         except tk.TclError:
             attr = ""
         return attr
@@ -411,16 +474,35 @@ class HtmlFrame(ttk.Frame):
 
     def scroll(self, event):
         "Handle mouse/touchpad scrolling"
-        if platform.system() == "Darwin":
-            self.html.yview_scroll(int(-1*(event.delta)), "units")
+        yview = self.html.yview()
+        if self.scroll_overflow and yview[0] == 0 and event.delta > 0:
+            self.scroll_overflow.scroll(event)
+        elif self.scroll_overflow and yview[1] == 1 and event.delta < 0:
+            self.scroll_overflow.scroll(event)
+        elif platform.system() == "Darwin":
+            self.html.yview_scroll(int(-1*event.delta), "units")
         else:
-            self.html.yview_scroll(int(-1*(event.delta)/40), "units")
+            self.html.yview_scroll(int(-1*event.delta/30), "units")
 
     def scroll_x11(self, event):
+        yview = self.html.yview()
         if event.num == 4:
-            self.html.yview_scroll(-4, "units")
+            if self.scroll_overflow and yview[0] == 0:
+                self.scroll_overflow.scroll_x11(event)
+            else:
+                self.html.yview_scroll(-4, "units")
         else:
-            self.html.yview_scroll(4, "units")
+            if self.scroll_overflow and yview[1] == 1:
+                self.scroll_overflow.scroll_x11(event)
+            else:
+                self.html.yview_scroll(4, "units")
+    
+    def overflow_scroll_x11(self, event):
+        yview = self.html.yview()
+        if event.num == 4 and self.scroll_overflow and yview[0] == 0:
+            self.scroll_overflow.scroll_x11(event)
+        elif self.scroll_overflow and yview[1] == 1:
+            self.scroll_overflow.scroll_x11(event)
 
     def load_html(self, html_source, base_url=None):
         "Reset parser and send html code to it"
@@ -430,17 +512,34 @@ class HtmlFrame(ttk.Frame):
             if not path.startswith("/"):
                 path = "/{0}".format(path)
             base_url = "file://{0}/".format(path)
-        self.current_url = base_url
-        self.html.base_url = base_url
+        self.html.base_url = self.current_url = base_url
         self.html.parse(html_source)
 
+        self.finish_css()
+
+    def finish_css(self):        
+        if self.waiting_for_reset:
+            self.waiting_for_reset = False
+            for style in self.accumulated_styles:
+                self.add_css(style)
+            self.accumulated_styles = []
+
     def add_html(self, html_source):
         "Parse HTML and add it to the end of the current document."
+        if not self.current_url:
+            path = currentpath(False)
+            if not path.startswith("/"):
+                path = "/{0}".format(path)
+            base_url = "file://{0}/".format(path)
+            self.html.base_url = self.current_url = base_url
         self.html.parse(html_source)
 
     def add_css(self, css_source):
         "Parse CSS code"
-        self.html.parse_css(data=css_source)
+        if self.waiting_for_reset:
+            self.accumulated_styles.append(css_source)
+        else:
+            self.html.parse_css(data=css_source, override=True)
 
 
 class HtmlLabel(HtmlFrame):
