@@ -1,3 +1,30 @@
+"""
+TkinterWeb v3.23
+This is a wrapper for the Tkhtml3 widget from http://tkhtml.tcl.tk/tkhtml.html, 
+which displays styled HTML documents in Tkinter.
+
+Copyright (c) 2024 Andereoo
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
+
+import mimetypes
 import sys, os
 import platform
 import threading
@@ -7,6 +34,8 @@ try:
 except ImportError:
     from urllib2 import Request, urlopen
 
+import ssl
+
 try:
     import tkinter as tk
     from tkinter import filedialog, ttk, colorchooser
@@ -381,6 +410,9 @@ INPUT[type="submit"],INPUT[type="button"], INPUT[type="reset"], BUTTON {
   color: #000000;
   color: tcl(::tkhtml::if_disabled #666666 #000000);
 }
+INPUT[disabled], BUTTON[disabled] {
+    cursor: auto;
+}
 INPUT[type="submit"]:after {
   content: "Submit";
 }
@@ -555,19 +587,39 @@ class AutoScrollbar(ttk.Scrollbar):
 class ScrolledTextBox(tk.Frame):
     "Text widget with a scrollbar."
 
-    def __init__(self, parent, **kwargs):
+    def __init__(self, parent, scroll_overflow=None, **kwargs):
 
         tk.Frame.__init__(self, parent)
 
+        self.scroll_overflow = scroll_overflow
+
         self.grid_rowconfigure(0, weight=1)
         self.grid_columnconfigure(0, weight=1)
         self.tbox = tbox = tk.Text(self, **kwargs)
-        tbox.grid(row=0, column=0, sticky="nsew", padx=2, pady=2)
+        tbox.grid(row=0, column=0, sticky="nsew")
 
         self.vsb = vsb = AutoScrollbar(self, command=tbox.yview)
         vsb.grid(row=0, column=1, sticky='nsew')
         tbox['yscrollcommand'] = self.check
 
+        tbox.bind("<MouseWheel>", self.scroll)
+        tbox.bind("<Button-4>", self.scroll_x11)
+        tbox.bind("<Button-5>", self.scroll_x11)
+
+    def scroll(self, event):
+        yview = self.tbox.yview()
+        if self.scroll_overflow and yview[0] == 0 and event.delta > 0:
+            self.scroll_overflow.scroll(event)
+        elif self.scroll_overflow and yview[1] == 1 and event.delta < 0:
+            self.scroll_overflow.scroll(event)
+
+    def scroll_x11(self, event):
+        yview = self.tbox.yview()
+        if event.num == 4 and self.scroll_overflow and yview[0] == 0:
+            self.scroll_overflow.scroll_x11(event)
+        elif event.num == 5 and self.scroll_overflow and yview[1] == 1:
+            self.scroll_overflow.scroll_x11(event)
+
     def check(self, *args):
         self.vsb.set(*args)
         if self.vsb.winfo_ismapped():
@@ -584,6 +636,9 @@ class ScrolledTextBox(tk.Frame):
     def configure(self, *args, **kwargs):
         self.tbox.configure(*args, **kwargs)
 
+    def insert(self, *args, **kwargs):
+        return self.tbox.insert(*args, **kwargs)
+
     def get(self, *args, **kwargs):
         return self.tbox.get(*args, **kwargs)
 
@@ -594,7 +649,7 @@ class ScrolledTextBox(tk.Frame):
 class FileSelector(tk.Frame):
     "File selector widget."
 
-    def __init__(self, parent, **kwargs):
+    def __init__(self, parent, accept, multiple, **kwargs):
         tk.Frame.__init__(self, parent)
         self.selector = selector = tk.Button(
             self, text="Browse", command=self.select_file)
@@ -603,56 +658,133 @@ class FileSelector(tk.Frame):
         selector.pack(side="left")
         label.pack(side="right", fill="both")
 
-    def select_file(self):
-        name = filedialog.askopenfilename()
-        if name:
-            name = name.replace("\\", "/").split("/")[-1]
-            self.label.config(text=name)
+        self.generate_filetypes(accept)
+
+        self.multiple = multiple
+        self.files = []
+    
+    def generate_filetypes(self, accept):
+        if accept:
+            accept_list = [ a.strip() for a in accept.split(",") ]
+            all_extensions = set()
+            filetypes = []
+
+            # First find all the MIME types
+            for mimetype in [ a for a in accept_list if not a.startswith(".") ]:
+                # the HTML spec specifies these three wildcard cases only:
+                if mimetype in ('audio/*', 'video/*', 'image/*'):
+                    extensions = [
+                        k for k, v in mimetypes.types_map.items()
+                        if v.startswith(mimetype[:-1])
+                    ]
+                else:
+                    extensions = mimetypes.guess_all_extensions(mimetype)
+                filetypes.append((mimetype, ' '.join(extensions)))
+                all_extensions.update(extensions)
+
+            # Now add any non-MIME types not already included as part of a MIME type.
+            for suffix in [ a for a in accept_list if a.startswith(".") ]:
+                if suffix not in all_extensions:
+                    mimetype = mimetypes.guess_type(f" {suffix}", suffix)[0]
+                    if mimetype:
+                        extensions = mimetypes.guess_all_extensions(mimetype)
+                        filetypes.append((mimetype, ' '.join(extensions)))
+                        all_extensions.update(extensions)
+                    else:
+                        filetypes.append((f"{suffix} files", suffix))
+
+            if len(filetypes) > 1:
+                filetypes.insert(0, ("All Supported Types", ' '.join(sorted(all_extensions))))
+
+            self.filetypes = filetypes
         else:
+            self.filetypes = []
+
+    def select_file(self):
+        if self.multiple:
+            self.files = files = filedialog.askopenfilenames(title="Select files", filetypes=self.filetypes)
+        else:
+            files = filedialog.askopenfilename(title="Select file", filetypes=self.filetypes)
+            if files:
+                self.files = files = (files,)
+        number = len(files)
+        if number == 0:
             self.label.config(text="No files selected.")
+        elif number == 1:
+            files = files[0].replace("\\", "/").split("/")[-1]
+            self.label.config(text=files)
+        else:
+            self.label.config(text="{} files selected.".format(number))
+        self.event_generate("<<Modified>>")
 
     def reset(self):
         self.label.config(text="No files selected.")
+        self.event_generate("<<Modified>>")
 
     def get_value(self):
-        text = self.label["text"]
-        if text == "No files selected.":
-            return ""
-        else:
-            return text
+        return self.files
 
     def configure(self, *args, **kwargs):
         self.selector.config(*args, **kwargs)
         if "activebackground" in kwargs:
             del kwargs["activebackground"]
         self.label.config(*args, **kwargs)
+        if "state" in kwargs:
+            del kwargs["state"]
         self.config(*args, **kwargs)
 
+
 class ColourSelector(tk.Frame):
     "Colour selector widget."
 
     def __init__(self, parent, colour, **kwargs):
-        tk.Button.__init__(self, parent, command=self.select_colour, bg="#ccc", activebackground="#aaa", width=5, highlightthickness=0, borderwidth=0)
+        tk.Frame.__init__(self, parent, bg="#ccc", highlightthickness=0)
         colour = colour if colour else "#000000"
         self.selector = tk.Button(self, bg=colour, command=self.select_colour, activebackground=colour, width=5, highlightthickness=0, borderwidth=0)
         self.selector.pack(expand=True, fill="both", padx=5, pady=5)
-        self.selector.bind("<Button-1>", lambda event: self.config(bg="#aaa"))
-        self.selector.bind("<ButtonRelease-1>", lambda event: self.config(bg="#ccc"))
+        self.bind("<Button-1>", self.on_click)
+        self.bind("<ButtonRelease-1>", self.on_release)
+        self.selector.bind("<Button-1>", self.on_click)
+        self.selector.bind("<ButtonRelease-1>", self.on_button_release)
         self.colour = colour
         self.default_colour = colour
+        self.disabled = True
+    
+    def on_release(self, event):
+        if not self.disabled:
+            self.config(bg="#ccc")
+            self.select_colour()
+    
+    def on_button_release(self, event):
+        if not self.disabled:
+            self.config(bg="#ccc")
+    
+    def on_click(self, event):
+        if not self.disabled:
+            self.config(bg="#aaa")
 
     def select_colour(self):
         colour = colorchooser.askcolor(title = "Choose color")[1]
         self.colour = colour if colour else self.colour
         self.selector.config(bg=self.colour, activebackground=self.colour)
+        self.event_generate("<<Modified>>")
+    
+    def configure(self, *args, **kwargs):
+        state = kwargs.pop("state")
+        if state == "disabled":
+            self.selector.config(state="disabled")
+            self.disabled = True
+        self.config(*args, **kwargs)
 
     def reset(self):
         self.colour = self.default_colour
         self.selector.config(bg=self.colour, activebackground=self.colour)
+        self.event_generate("<<Modified>>")
 
     def get_value(self):
         return self.colour
 
+
 class Notebook(ttk.Frame):
     """Drop-in replacement for the ttk.Notebook widget,
     The ttk.Notebook widget is incompatable with Tkhtml on some platforms, causing crashes when selecting tabs
@@ -763,7 +895,7 @@ class PlaceholderThread:
         return True
 
 
-def download(url, data=None, method="GET", decode=None):
+def download(url, data=None, method="GET", decode=None, insecure=False):
     "Fetch files."
     "Technically this isn't thread-safe (even though it is being used inside threads by Tkinterweb, "
     "but as long as install_opener() is not called and a string is used as the url parameter we are okay."
@@ -771,12 +903,17 @@ def download(url, data=None, method="GET", decode=None):
     if url in BUILTINPAGES:
         return BUILTINPAGES[url], url, 'text/html'
 
+    ctx = ssl.create_default_context()
+    if insecure:
+        ctx.check_hostname = False
+        ctx.verify_mode = ssl.CERT_NONE
+
     thread = threadname()
     url = url.replace(" ", "%20")
     if data and (method == "POST"):
-        req = urlopen(Request(url, data, headers=HEADER))
+        req = urlopen(Request(url, data, headers=HEADER), context=ctx)
     else:
-        req = urlopen(Request(url, headers=HEADER))
+        req = urlopen(Request(url, headers=HEADER), context=ctx)
     if not thread.isrunning():
         return None, url, ""
     data = req.read()
@@ -826,13 +963,24 @@ def currentpath(use_file=True):
     else:
         return os.getcwd()
 
+
 def strip_css_url(url):
     return url[4:-1].replace("'", "").replace('"', '')
 
+
+def get_fonts():
+    return os.path.join(currentpath(), "tkhtml", "opensans.ttf")
+
+
 def get_tkhtml_folder():
-    return os.path.join(currentpath(), "tkhtml",
-                        platform.system(),
-                        "64-bit" if sys.maxsize > 2**32 else "32-bit")
+    if platform.system() == "Linux" or platform.system() == "Darwin":
+        if "arm" in os.uname()[-1]: # for Raspberry Pi (#24) and M1 Mac (#26)
+            bit = "arm"
+        else:
+            bit = "64-bit" if sys.maxsize > 2**32 else "32-bit"
+    else:
+        bit = "64-bit" if sys.maxsize > 2**32 else "32-bit"
+    return os.path.join(currentpath(), "tkhtml", platform.system(), bit)
 
 
 def load_tkhtml(master, location=None, force=False):
@@ -840,8 +988,7 @@ def load_tkhtml(master, location=None, force=False):
     global tkhtml_loaded
     if (not tkhtml_loaded) or force:
         if location:
-            path = ["linsert", "$auto_path", 0, location]
-            master.tk.call("set", "auto_path", path)
+            master.tk.eval("set auto_path [linsert $auto_path 0 {"+location+"}]")
         master.tk.eval("package require Tkhtml")
         tkhtml_loaded = True
 
@@ -852,7 +999,7 @@ def load_combobox(master, force=False):
     if not (combobox_loaded) or force:
         path = os.path.join(currentpath(), "tkhtml")
         master.tk.call("lappend", "auto_path", path)
-        master.tk.eval("package require combobox")
+        master.tk.call("package", "require", "combobox")
         combobox_loaded = True
 
 
@@ -863,3 +1010,10 @@ def notifier(text):
     except Exception:
         "sys.stdout.write doesn't work in .pyw files."
         "Since .pyw files have no console, we can simply not bother printing messages."
+
+def tkhtml_notifier(name, text, *args):
+    "Tkhtml -logcmd printer."
+    try:
+        sys.stdout.write("DEBUG "+str(name)+": "+str(text)+"\n\n")
+    except Exception:
+        pass
